using System.Net.Mime;
using System.Text;
using FitConnect;
using FitConnect.Models;
using FitConnect.Models.Api.Metadata;
using FitConnect.Models.Api.Set;
using FluentAssertions;
using Microsoft.Extensions.Logging;
using Attachment = FitConnect.Models.Attachment;

namespace E2ETest;

[Order(20)]
[TestFixture]
public class StraightForwardTest : EndToEndTestBase {
    private SentSubmission _submission = null!;
    const string REPLY_CHANNEL_EMAIL = "dummy@trashmail.com";
    private const string TRANSACTION_ID = "9xxd-432x-6543-xfd6-gfdx-fd28";
    private const double GROSS_AMOUNT = 12.5;
    private const string AUTHENTICATION_VERSION = "1.0.34";

    private const string AUTHENTICATION_CONTENT =
        "QXV0b21hdGljYWxseSBCYXNlNjRVcmwgZW5jb2RlZCBTdHJpbmc";

    [Order(10)]
    [Test]
    public void SendingSubmission() {
        var outgoing = SendableSubmission.Builder()
            .SetDestination(Settings.DestinationId)
            .SetServiceType(Settings.ServiceIdentifier, "Straight forward test")
            .SetJsonData(@"{""data"":""value""}")
            .AddAttachments(Attachment.FromPath("Test.pdf","application/pdf", description: "Attachment #1", fileName: "Test.pdf"),
                Attachment.FromPath("Test.pdf", "application/pdf", description:"Attachment #2", fileName: "Test.pdf"))
            .AddAttachment(Attachment.FromPath("Test.pdf", "application/pdf", description:"Attachment #3", fileName: "Test.pdf"))
            .AddAttachment(Attachment.FromByteArray(
                Encoding.UTF8.GetBytes("{\"message\":\"Hello World\"}"),
                MediaTypeNames.Application.Json, "JsonFileAttachment.json", "Attachment #4"))
            .SetAuthenticationInformation(new AuthenticationInformation(
                AUTHENTICATION_CONTENT,
                AuthenticationInformationType.IdentificationReport,
                AUTHENTICATION_VERSION))
            .SetPaymentInformation(new PaymentInformation(
                PaymentMethod.Invoice,
                PaymentStatus.Booked,
                TRANSACTION_ID,
                "E2E Testing the payment information"
            ) {
                GrossAmount = GROSS_AMOUNT
            })
            .SetReplyChannel(new EMail(REPLY_CHANNEL_EMAIL))
            .Build();

        _submission = Sender.SendAsync(outgoing).Result;

        _submission.Should().NotBeNull();


        _submission.AuthenticationTags.Metadata.Should().NotBeNullOrWhiteSpace();
        _submission.AuthenticationTags.Data.Should().NotBeNullOrWhiteSpace();
        _submission.AuthenticationTags.Attachments.Should().NotBeNullOrEmpty();
        _submission.AuthenticationTags.Attachments!.All(item =>
                !(string.IsNullOrWhiteSpace(item.Value) ||
                  string.IsNullOrWhiteSpace($"{item.Key}")))
            .Should().BeTrue();
    }

    [Order(11)]
    [Test]
    public void SendingSubmission_WithoutAttachments() {
        var outgoing = SendableSubmission.Builder().SetDestination(Settings.DestinationId)
            .SetServiceType(Settings.ServiceIdentifier, "Straight forward test")
            .SetJsonData(@"{""data"":""value""}")
            .Build();

        var submission = Sender.SendAsync(outgoing).Result;

        var caseId = submission.CaseId;
        var submissionId = submission.SubmissionId;
    }

    [Test]
    [Order(20)]
    public void Sender_GetSubmissionState() {
        // Act
        var status = Sender.GetStatusForSubmissionAsync(_submission)
            .Result;

        // Assert
        status.SubmissionState.Should().Be(EventType.Submit);
    }

    [Test]
    [Order(21)]
    public void Sender_GetEventLogAfterSubmit() {
        // Act
        var status = Sender.GetEventLogAsync(_submission.CaseId, Settings.DestinationId).Result;

        // Assert
        status.Count.Should().BeGreaterThan(0);
        status.ForEach(
            s => Logger.LogInformation("Status {When} {Event}", s.EventTime, s.EventType));
    }

    [Test]
    [Order(30)]
    public void Subscriber_GetSubmissionState() {
        // Act
        var status = Subscriber.GetEventLogAsync(_submission.CaseId, Settings.DestinationId).Result;

        // Assert
        status.Count.Should().BeGreaterThan(0);
        status.ForEach(
            s => Logger.LogInformation("Status {When} {Event}", s.EventTime, s.EventType));
    }

    [Test]
    [Order(40)]
    public void RequestSubmission() {
        var receivedSubmission =
            Subscriber.RequestSubmissionAsync(_submission.SubmissionId).Result;

        var data = receivedSubmission.GetData();
        Logger.LogInformation("Data {Data}", data);

        data.Should().Be(@"{""data"":""value""}");

        var attachments = receivedSubmission.GetAttachmentsAsync().Result.ToList();
        attachments.Count(a => a.Filename == "Test.pdf").Should().Be(3);
        attachments.TrueForAll(a => a.Description?.StartsWith("Attachment #") ?? false).Should()
            .BeTrue();
        attachments.Count.Should().Be(4);
        var size = File.ReadAllBytes("Test.pdf").Length;
        attachments.Count(a => a.GetBytes().Length == size).Should().Be(3);

        attachments.Single(attachment => attachment.Filename == "JsonFileAttachment.json").GetData()
            .Should().BeEquivalentTo("{\"message\":\"Hello World\"}");

        receivedSubmission.SubmissionId.Should().Be(_submission.SubmissionId);
        receivedSubmission.CaseId.Should().Be(_submission.CaseId);
        receivedSubmission.GetDataMimeType().Should()
            .Be(MediaTypeNames.Application.Json);

        receivedSubmission.Metadata.Should().NotBeNull();

        var authenticationInformation =
            receivedSubmission.Metadata?.AuthenticationInformation?.First();
        authenticationInformation.Should().NotBeNull();
        authenticationInformation!.Type
            .Should().Be(AuthenticationInformationType.IdentificationReport);
        authenticationInformation!.Version.Should().Be(AUTHENTICATION_VERSION);
        authenticationInformation!.Content.Should().Be(AUTHENTICATION_CONTENT);


        receivedSubmission.Metadata?.PaymentInformation.GrossAmount.Should().Be(GROSS_AMOUNT);
        receivedSubmission.Metadata?.PaymentInformation.TransactionId.Should()
            .Be(TRANSACTION_ID);
        receivedSubmission.Metadata?.PaymentInformation.PaymentMethod.Should()
            .Be(PaymentMethod.Invoice);

        receivedSubmission.Metadata?.ReplyChannel?.EMail?.Address.Should().Be(REPLY_CHANNEL_EMAIL);

        receivedSubmission.AcceptSubmissionAsync().Wait();
    }

    [Test]
    [Order(50)]
    public void Sender_GetEventLog_AfterAccepting() {
        // Act
        var status = Sender.GetEventLogAsync(_submission.CaseId, Settings.DestinationId).Result;

        // Assert
        status.Count.Should().BeGreaterThan(0);
        status.Any(s => s.EventType == EventType.Accept);

        var submitStatus = status.Single(s => s.EventType == EventType.Submit);
        submitStatus.Events?.SubmitSubmissionEvent?.AuthenticationTags?.Attachments.Should()
            .NotBeNullOrEmpty();

        status.ForEach(
            s => Logger.LogInformation("Status {When} {Event}", s.EventTime, s.EventType));
    }

    [Test]
    [Order(51)]
    public void Sender_GetSubmissionState_AfterAccepting() {
        var status = Sender.GetStatusForSubmissionAsync(_submission).Result;

        status.SubmissionState.Should().Be(EventType.Accept);
        status.Problems.Should().BeNullOrEmpty();
    }

    [Test]
    [Order(60)]
    public void Sender_GetStatus_AfterAccepting() {
        // Act
        var (status, problems) =
            Sender.GetStatusForSubmissionAsync(_submission).Result;

        // Assert
        status.Should().Be(EventType.Accept);
    }

    [Test]
    [Order(61)]
    public void Sender_GetStatusForSubmission_FailsOnManipulatedAuthenticationTags() {
        var manipulated = new SentSubmission(_submission.SubmissionId, _submission.CaseId,
            _submission.DestinationId,
            new AuthenticationTags() {
                Data = _submission.AuthenticationTags.Data + "Invalid",
                Metadata = _submission.AuthenticationTags.Metadata + "Invalid",
                Attachments = _submission.AuthenticationTags.Attachments!.ToDictionary(
                    i => i.Key,
                    i => "InvalidTag!")
            });

        var exception = Assert.Throws<AggregateException>(() =>
            Sender.GetStatusForSubmissionAsync(manipulated).Wait());

        exception.InnerExceptions.Any(e => e is AuthenticationTagException).Should().BeTrue();
    }
}
